/******************************************************************
 A simple HBF 'API' by Yidao Cai
 October 12, 1993

 This 'API' (if it can be called an API at all) preforms the basic 
 function of getting the bitmap of a character.  It is intended to 
 users who prefer an API with simpler functions and less codes.  It
 should work on all platforms with minimum modification.

 It is free and can be modified to fit your needs, but I expect you 
 properly credit my work.

 Old functional protocal is used.  Change if your compiler doesn't like it.

 Send comments and report bugs to cai@neurophys.wisc.edu

You must call HBFopen() before call HBFgetBitmap(), and you need
to call HBFclose() before you make another HBFopen() call unless
the previous HBFopen() call is unsuccessful.  A sample application
is enclosed at the end of this file.

int HBFopen(char *hbffile)
	Open HBF file (hbffile) and bitmap files, prepare storage space 
	(buffer) for subsequent call of HBFgetBitmap(), get sizes (mx, my 
	and CHsize) of bitmap.  The HBF file name (hbffile) is absolute
	while the bitmap file name is relative to the 'root', ie. it uses
	'root' + 'bitmap-file' to open a bitmap file.

int HBFgetBitmap(unsigned int code)
	Get the bitmap of character 'code' and store it in 'buffer'.

void HBFclose()
	Close HBF, clear storage.

Return values 
	1:	normal
	-2:	can't open HBF file
	-3:	can't open bitmap file
	-4:	out of range character
	-5:	more than 2 byte-2-ranges
	-6:	Not enough memory

DOS: you may need to change calloc to farcalloc, free to farfree, or whatever
     appropriate, depending on the memory model you use.
*******************************************************************/  

/**********************************************************************
   Modified by Tian Boagng,    November 1993
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#define PUBLIC
#define PRIVATE static

PUBLIC int mx, my, yd;          /* width, height and descent of bitmap */
PUBLIC int CHsize;              /* size of bitmap, or length of buffer */
PUBLIC unsigned char *buffer;   /* store the bitmap of each char, updated on
                                   each call of HBFgetBitmap() */

PRIVATE struct Bytes {unsigned int l; unsigned int h; unsigned int a;} *b2;
PRIVATE struct HBFrange {unsigned int l; unsigned int h; long offset;
        char bmfname[50]; FILE *fp; short sideway; } *seg;
PRIVATE int nb2, nseg;
PRIVATE FILE *HBF;

PRIVATE long *SegLength, maxb2Length, codeIndex = 0;
PRIVATE int segIndex = 0;
extern  short SideWay;
PUBLIC  void HBFclose();

PUBLIC int HBFopen(hbffile)
char *hbffile;
{
  int j, k, m, xd;
  char cst[120], buf[10];

  if ((HBF=fopen(hbffile, "r"))==NULL) {
    fprintf(stderr, "Can't open HBF file -> %s\n", hbffile);
    return -2;
  }

  while (fgets(cst, 119, HBF)!=NULL)
    if (strncmp(cst, "HBF_BITMAP_BOUNDING_BOX", 23)==0) break;
  sscanf(cst, "%*s%d%d%d%d", &mx, &my, &xd, &yd);
  CHsize=my*((mx+7)/8);
  if (buffer!=NULL) free(buffer);
  buffer = (unsigned char *) calloc(CHsize+1, sizeof(unsigned char));
  if (buffer==NULL) {
    printf("Not enough memory\n"); return -6;
  }

  while (fgets(cst, 119, HBF)!=NULL)
    if (strncmp(cst, "HBF_START_BYTE_2_RANGES", 23)==0) break;
  sscanf(cst, "%*s%d", &nb2);
  b2 = (struct Bytes *) calloc(nb2, sizeof(struct Bytes));
  if (b2==NULL) {
    printf("Not enough memory\n"); return -6;
  }
  maxb2Length = 0L;
  for (j=0; j<nb2; j++) {
    while (fgets(cst, 119, HBF)!=NULL)
      if (strncmp(cst, "HBF_BYTE_2_RANGE", 16)==0) break;
    sscanf(cst, "%*s%i-%i", &(b2[j].l), &(b2[j].h));
    b2[j].a = b2[j].h - b2[j].l +1;
    maxb2Length += (long) b2[j].a;
  }

  while (fgets(cst, 119, HBF)!=NULL)
    if (strncmp(cst, "HBF_START_CODE_RANGES", 21)==0) break;
  sscanf(cst, "%*s%d", &nseg);
  seg = (struct HBFrange *) calloc(nseg, sizeof(struct HBFrange));
  if (seg==NULL) {
    printf("Not enough memory\n"); return -6;
  }
  SegLength = (long *) calloc(nseg, sizeof(long));
  if (SegLength==NULL) {
    printf("Not enough memory\n"); return -6;
  }
  for (j=0; j<nseg; j++) {
    while (fgets(cst, 119, HBF)!=NULL)
      if (strncmp(cst, "HBF_CODE_RANGE", 14)==0) break;
    sscanf(cst, "%*s%i-%i%s%li %s", &(seg[j].l), &(seg[j].h),
           seg[j].bmfname, &(seg[j].offset), buf);
    if (j>0) {
      for (m=0; m<j; m++)
      if (strcmp(seg[m].bmfname, seg[j].bmfname)==0)
        seg[j].fp=seg[m].fp;
    }
    if (seg[j].fp==NULL) {
      strcpy(cst, seg[j].bmfname);
      if ((seg[j].fp = fopen(cst, "rb"))==NULL) {
        fprintf(stderr, "Can't open bitmap font file -> %s\n", cst);
        return -3;
      }
    }
    seg[j].sideway = buf[0]=='s' || buf[0]=='S' ;
    SegLength[j] = ((seg[j].h >> 8) - (seg[j].l >> 8) + 1)*maxb2Length
                   - (b2[nb2-1].h - (seg[j].h & 0x00FF))
                   - ((seg[j].l & 0x00FF) - b2[0].l) ;
  }
  return 1;
}

PUBLIC void HBFclose()
{
  int j;

  fclose(HBF);
  for (j=0; j<nseg; j++) fclose(seg[j].fp);
  free(seg); free(b2);
  free (SegLength);
}
 
PUBLIC int HBFgetNextBitmap ()
{
  int ret;

  if (segIndex >= nseg) return 0;
  if (codeIndex == 0) {
    fseek(seg[segIndex].fp, seg[segIndex].offset, 0);
    SideWay = seg[segIndex].sideway;
  }
  ret = (int) fread(buffer, CHsize, 1, seg[segIndex].fp);
  if (ret > 0) {
    codeIndex++;
    if (codeIndex >= SegLength[segIndex]) {segIndex++;  codeIndex = 0;}
  }
  return (ret);
}
